"""
示例PU单元,调用其他单元及全局服务单元接口
"""
from dalpyframe.unitbase import ProcessUnit
class CallUnit(ProcessUnit):
    """
    常驻任务单元,需描述最小资源要求

    Attributes:
        CPU: 100
        Memory: 64
        Version: 1.0.1
    """
    def __init__(self):
        super().__init__()
        self._h_usercenter_sys = self.register_request('pu_sys_usercenter','1.0.1','system')
        self._h_glb_unitcenter_sys = self.register_glb_request('pu_sys_unitcenter','1.0.1','system')
        self._h_glb_unitcenter_copy_create = self.register_glb_request('pu_sys_unitcenter',
            '1.0.1','copy_create')
        self._h_glb_unitcenter_copy_del = self.register_glb_request('pu_sys_unitcenter',
            '1.0.1','copy_delete')
        self._h_glb_unitcenter_copy_list = self.register_glb_request('pu_sys_unitcenter',
            '1.0.1','copy_list')

    def custom_init(self)->bool:
        """
        业务单元自定义初始化
        """
        return True

    def example1_data(self,_caller,_index):
        """
        返回其他单元接口返回值

        Attributes:
            ACCode: 111
        Returns:
            array (array): 返回值
        Examples:
            {}
        """
        ret = self.call_unit(self._h_usercenter_sys,{'oper':'query_info'})
        return ret

    def example2_data(self,_caller,_index):
        """
        返回全局基础服务单元接口system返回值

        Attributes:
            ACCode: 111
        Returns:
            array (array): 返回值
        Examples:
            {}
        """
        ret = self.call_glb_unit(self._h_glb_unitcenter_sys,{'oper':'query_info'})
        return ret

    def example_copy_create_data(self,_caller,_index):
        """
        创建已运行单元copy实例

        Attributes:
            ACCode: 111
        Returns:
            array (array): 返回值
        Examples:
            {}
        """
        ret = self.call_glb_unit(self._h_glb_unitcenter_copy_create,
            {'name':'pu_example_store','version':'1.0.1','tag':'hello1'})
        return ret

    def example_copy_delete_data(self,_caller,_index):
        """
        删除单元copy实例

        Attributes:
            ACCode: 111
        Returns:
            array (array): 返回值
        Examples:
            {}
        """
        ret = self.call_glb_unit(self._h_glb_unitcenter_copy_del,
            {'name':'pu_example_store','version':'1.0.1','tag':'hello1'})
        return ret

    def example_copy_list_data(self,_caller,_index):
        """
        单元copy实例列表

        Attributes:
            ACCode: 111
        Returns:
            array (array): 返回值
        Examples:
            {}
        """
        ret = self.call_glb_unit(self._h_glb_unitcenter_copy_list,
            {'name':'pu_example_store','version':'1.0.1'})
        return ret

    def test(self):
        """
        单元测试
        """
        self.example1_data({},{})
        return True
