"""
示例PU单元,仅提供一个数据调用接口和对应的数据订阅接口,返回随机字符串,用于演示框架使用
"""
import json
from sqlalchemy import Column
from sqlalchemy import Integer
from sqlalchemy import String
from sqlalchemy import text
from faker import Faker
from dalpyframe.unitbase import ProcessUnit, db_session, Base

# pylint: disable=too-few-public-methods


class User(Base):
    """
    用户类,对应mysql数据中用户表
    """
    __tablename__ = 'users'
    id = Column(Integer, primary_key=True)
    name = Column(String(50, collation='utf8mb4_unicode_ci'))
    fullname = Column(String(50, collation='utf8mb4_unicode_ci'))
    nickname = Column(String(50, collation='utf8mb4_unicode_ci'))

    def __repr__(self):
        return f"<User(name='{self.name}', fullname='{self.fullname}', nickname='{self.nickname}')>"


class DBExample(ProcessUnit):
    """
    常驻任务单元,需描述最小资源要求

    Attributes:
        CPU: 100
        Memory: 64
        Version: 1.0.1
    """

    def __init__(self):
        super().__init__()
        self._faker = Faker(locale='zh_CN')

    def custom_init(self) -> bool:
        """
        业务单元自定义初始化
        """
        self.logger.info("sql config is:%s", self.sql_config)
        return True

    @db_session
    def insert_data(self, _caller, _index):
        """
        获取模拟客户信息

        Attributes:
            ACCode: 111
        Returns:
            {}
        Examples:
            {}
        """
        name = self._faker.name()
        # ORM用法
        ed_user = User(name=name, fullname=name, nickname=name[1:])
        session.add(ed_user)
        session.commit()
        # Core用法
        result = self.sql_connection.execute(text("select * from users"))
        for row in result:
            self.logger.info(row)
        return (0, 'success', {})

    def test(self):
        """
        单元测试
        """
        self.insert_data({}, {})
        return True
