"""
示例PU单元,仅提供一个数据调用接口和对应的数据订阅接口,返回随机字符串,用于演示框架使用
"""
import threading
from datetime import datetime
from faker import Faker
from dalpyframe.unitbase import ProcessUnit


class FakeIF(ProcessUnit):
    """
    常驻任务单元,需描述最小资源要求

    Attributes:
        CPU: 100
        Memory: 64
        Version: 1.0.1
        BaseImg: python:3.9.16-alpine
    """

    def __init__(self):
        super().__init__()
        self._faker = Faker(locale='zh_CN')
        self._ds_notify = {
            'example': {
                'mutex': threading.Lock(),
                'dt_updated': datetime.now(),
                'dt_notified': datetime.now()}}

    def custom_init(self) -> bool:
        """
        业务单元自定义初始化
        """
        return True

    def get_example_data(self, _caller, _index):
        """
        获取模拟客户信息

        Attributes:
            ACCode: 111
        Args:
            name (str, optional): 随便一个名字,演示可选入参写法
        Returns:
            name (str): 客户名
            ssn (str): 身份证号码
            mobile (str): 手机号码
            company (str): 公司名
            email (str): 电子邮箱
            address (str): 地址
        Examples:
            {}
        """
        fake_data = {}
        fake_data['name'] = self._faker.name()
        fake_data['ssn'] = self._faker.ssn()
        fake_data['mobile'] = self._faker.phone_number()
        fake_data['company'] = self._faker.company()
        fake_data['email'] = self._faker.email()
        fake_data['address'] = self._faker.address()
        return (0, 'success', fake_data)

    def new_example_data(self, _caller, _index):
        """
        新模拟客户信息

        Attributes:
            ACCode: 111
        Args:
            name (str, optional): 随便一个名字,演示可选入参写法
        Returns:
            {}
        Examples:
            {}
        """
        with self._ds_notify['example']['mutex']:
            self._ds_notify['example']['dt_updated'] = datetime.now()
        return (0, 'success', {})

    def example_cron(self):
        """
        数据example的变更推送，默认秒级，高频数据推送可改为thread方式
        默认为全量提取推送，需做性能优化时，按数据性质实现单条数据的new,upd,del指令，并在使用处合成完整数据

        Attributes:
            cron: * * * * * */1
        Returns:
            example
        """
        ret_data = []
        with self._ds_notify['example']['mutex']:
            if self._ds_notify['example']['dt_notified'] >= \
                    self._ds_notify['example']['dt_updated']:
                return ret_data
        self._ds_notify['example']['dt_notified'] = datetime.now()
        sub_idx = self.get_sub_byds('example')
        for item in sub_idx:
            ret_code = 1
            if item[1] == 'get_example':
                (ret_code, _, ret_content) = self.get_example_data(
                    item[2], item[3])
            if ret_code == 0:
                ret_data.append((item, 'set', ret_content))
        return ret_data

    def test(self):
        """
        单元测试
        """
        self.get_example_data({}, {})
        return True
