"""
示例PU单元,订阅其他单元接口,接收数据更新
"""
import json
from dalpyframe.unitbase import ProcessUnit


class SubUnit(ProcessUnit):
    """
    在test产品环境中测试使用

    Attributes:
        CPU: 100
        Memory: 64
        Version: 1.0.1
    """

    def __init__(self):
        super().__init__()
        self._h_ws_example1 = self.register_subscription('pu_example_ws',
                                                         '1.0.1', 'example1', self._exmaple_update)
        self._h_ws_example2 = self.register_subscription('pu_example_ws',
                                                         '1.0.1', 'example2', self._exmaple_update)

    def custom_init(self) -> bool:
        """
        业务单元自定义初始化
        """
        return True

    def _exmaple_update(self, interface, index, data):
        self.logger.info("update from %s index %s data:%s",
                         interface, index, data)

    def subscribe_data(self, _caller, index):
        """
        返回其他单元接口返回值

        Attributes:
            ACCode: 111-
        Returns:
            array (array): 返回值
        Examples:
            {}
        """
        if index['target'] == 'example1':
            self.subscribe(self._h_ws_example1, {})
        elif index['target'] == 'example2':
            self.subscribe(self._h_ws_example2, {})
        elif index['target'] == 'example_index':
            self.subscribe(self._h_ws_example1, json.loads(index['index']))
        return (0, 'success', {})

    def unsubscribe_data(self, _caller, index):
        """
        返回其他单元接口返回值

        Attributes:
            ACCode: 111
        Returns:
            array (array): 返回值
        Examples:
            {}
        """
        if index['target'] == 'example1':
            self.unsubscribe(self._h_ws_example1, {})
        elif index['target'] == 'example2':
            self.unsubscribe(self._h_ws_example2, {})
        elif index['target'] == 'example_index':
            self.unsubscribe(self._h_ws_example1, json.loads(index['index']))
        return (0, 'success', {})

    @classmethod
    def test(cls):
        """
        单元测试
        """
        return True
