"""
示例PU单元,提供一个调用耗时较长的接口,演示框架内置的异步任务用法
"""
from faker import Faker
from dalpyframe.unitbase import ProcessUnit


class TaskCall(ProcessUnit):
    """
    常驻任务单元,需描述最小资源要求

    Attributes:
        CPU: 100
        Memory: 64
        Version: 1.0.1
    """

    def __init__(self):
        super().__init__()
        self._faker = Faker(locale='zh_CN')

    def custom_init(self) -> bool:
        """
        业务单元自定义初始化
        """
        return True

    def example_data(self, _caller, _index):
        """
        获取模拟客户信息

        Attributes:
            ACCode: 111
        Args:
            name (str, optional): 随便一个名字,演示可选入参写法
        Returns:
            name (str): 客户名
            ssn (str): 身份证号码
            mobile (str): 手机号码
            company (str): 公司名
            email (str): 电子邮箱
            address (str): 地址
        Examples:
            {}
        """
        fake_data = {}
        for i in range(1, 10):
            self.task_update_progress(i*10)
        fake_data['name'] = self._faker.name()
        fake_data['ssn'] = self._faker.ssn()
        fake_data['mobile'] = self._faker.phone_number()
        fake_data['company'] = self._faker.company()
        fake_data['email'] = self._faker.email()
        fake_data['address'] = self._faker.address()
        return (0, 'success', fake_data)

    def test(self):
        """
        单元测试
        """
        self.example_data({}, {})
        return True
