"""
示例PU单元,提供两个可订阅接口,返回随机字符串,演示框架数据订阅使用
"""
from random import randint
import time
from faker import Faker
from dalpyframe.unitbase import ProcessUnit


class WSExample(ProcessUnit):
    """
    常驻任务单元,需描述最小资源要求

    Attributes:
        CPU: 100
        Memory: 1024000
        Version: 1.0.1
        _faker (Faker): 假数据生成器
        _fake_data1 (dic): 类型1假数据
        _date_time1 (time): 数据生成时间
        _fake_data2 (dic): 类型1假数据
        _date_time2 (time): 数据生成时间
    """

    def __init__(self):
        super().__init__()
        self._faker = Faker(locale='zh_CN')
        self._fake_data1 = {}
        self._fake_data2 = {}
        self._date_time1 = None
        self._date_time2 = None

    @classmethod
    def custom_init(cls):
        """
        业务单元自定义初始化
        """
        return True

    def example1_data(self, _caller, _index):
        """
        获取模拟客户信息

        Attributes:
            ACCode: 111
        Returns:
            name(str): 客户名
            ssn(str): 身份证号码
            mobile(str): 手机号码
            company(str): 公司名
            email(str): 电子邮箱
            address(str): 地址
        Examples:
            {}
        """
        return (0, 'success', [(self._date_time1, self._fake_data1)])

    def example2_data(self, _caller, _index):
        """
        获取模拟客户信息

        Attributes:
            ACCode: 111
        Returns:
            name(str): 客户名
            ssn(str): 身份证号码
            mobile(str): 手机号码
            company(str): 公司名
            email(str): 电子邮箱
            address(str): 地址
        Examples:
            {}
        """
        return (0, 'success', [(self._date_time2, self._fake_data2)])

    def maintain_cron(self):
        """
        秒级及以上定时任务,用于维护数据,线程数量可以根据业务逻辑自由安排,可以一个单元一个,也可以一个接口一个,使用*_cron声明即可;
        频率高于秒级使用*_thread;
        返回格式为{接口名:[(变更类型,索引,数据变更时间戳,最新数据),...],...}
        此示例中接口名为example
        变更类型为new del update 之一
        索引为数据在接口中的索引值,根据业务逻辑自定义,所有数据要求有索引,单值数据索引可以为空
        最新数据为变更后的最新数据,如为del操作,最新数据为空

        Attributes:
            cron: * * * * * */1

        Returns:
            example1 example2
        """
        ret_data = {}
        rand_index = [{}, {'market': 'sh'}, {
            'market': 'sh', 'code': '600030'}, {'y': 1}]
        ret_data['example1'] = []
        self._fake_data1['name'] = self._faker.name()
        self._fake_data1['ssn'] = self._faker.ssn()
        self._fake_data1['mobile'] = self._faker.phone_number()
        self._fake_data1['company'] = self._faker.company()
        self._fake_data1['email'] = self._faker.email()
        self._fake_data1['address'] = self._faker.address()
        self._date_time1 = time.localtime()
        index = rand_index[randint(0, len(rand_index)-1)]
        ret_data['example1'].append(
            ('update', index, self._date_time1, self._fake_data1))

        ret_data['example2'] = []
        self._fake_data2['name'] = self._faker.name()
        self._fake_data2['ssn'] = self._faker.ssn()
        self._fake_data2['mobile'] = self._faker.phone_number()
        self._fake_data2['company'] = self._faker.company()
        self._fake_data2['email'] = self._faker.email()
        self._fake_data2['address'] = self._faker.address()
        self._date_time2 = time.localtime()
        index = rand_index[randint(0, len(rand_index)-1)]
        ret_data['example2'].append(
            ('update', index, self._date_time2, self._fake_data2))
        return ret_data

    def maintain_thread(self):
        """
        常驻线程,用于维护数据,线程数量可以根据业务逻辑自由安排,可以一个单元一个,也可以一个接口一个,使用*_thread声明即可;
        刷新频率较低的数据维护任务使用*_cron
        返回格式为[(接口名,变更类型,索引,数据变更时间戳,最新数据),...]
        此示例中接口名为example
        变更类型为new del update 之一
        索引为数据在接口中的索引值,根据业务逻辑自定义,所有数据要求有索引,单值数据索引可以为空
        最新数据为变更后的最新数据,如为del操作,最新数据为空

        Returns:
            example2
        """
        ret_data = {}
        rand_index = [{}, {'market': 'sh'}, {
            'market': 'sh', 'code': '600030'}, {'y': 1}]
        ret_data['example2'] = []
        self._fake_data2['name'] = self._faker.name()
        self._fake_data2['ssn'] = self._faker.ssn()
        self._fake_data2['mobile'] = self._faker.phone_number()
        self._fake_data2['company'] = self._faker.company()
        self._fake_data2['email'] = self._faker.email()
        self._fake_data2['address'] = self._faker.address()
        self._date_time2 = time.localtime()
        index = rand_index[randint(0, len(rand_index)-1)]
        ret_data['example2'].append(
            ('update', index, self._date_time2, self._fake_data2))
        time.sleep(1)
        return ret_data

    @classmethod
    def test(cls):
        """
        单元测试
        """
        return True
