"""
爬取期货业协会,期货公司信息列表,每日执行
"""
import json
import time
import random
from sqlalchemy import Column
from sqlalchemy import Integer
from sqlalchemy import String
from sqlalchemy.dialects.mysql import insert
import requests
from dalpyframe.unitbase import TaskUnit
from dalpyframe.frame import Base
# pylint:disable=too-few-public-methods
class FutuID(Base):
    """
    基础信息
    """
    __tablename__ = 'cfa_china_furcomp_id'
    compId = Column(Integer,primary_key=True)
    organId = Column(String(12,collation='utf8mb4_unicode_ci'))
    name = Column(String(16,collation='utf8mb4_unicode_ci'))
    def __repr__(self):
        return f"<FutuID(name='{self.name}'),compId='{self.compId}',organId='{self.organId}'>"
class CFAChina(TaskUnit):
    """
    常驻计划任务单元,需描述最小资源要求和计划安排

    Attributes:
        CPU: 100
        Memory: 64
        Version: 1.0.1
        Cron: */1 * * * *
    """
    def req_json(self,url,payload=None)->tuple:
        """
        封装post请求
        """
        header = {
            "Pragma":"no-cache",
            "Accept":"application/json, text/javascript, */*; q=0.01",
            "Cache-Control":"no-cache",
            "X-Requested-With":"XMLHttpRequest",
            "Content-Type":"application/x-www-form-urlencoded; charset=UTF-8",
            "User-Agent":
        "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) \
Chrome/72.0.3626.81 Safari/537.36"}
        if payload:
            resp = requests.post(url,data=json.dumps(payload),headers=header,timeout=30)
        else:
            resp = requests.post(url,headers=header,timeout=30)
        time.sleep(random.uniform(1,3))
        if resp.status_code != 200:
            self.logger.warn('request [%s] failed, code[%d], info[%s]',
            url,resp.status_code,resp.text)
        return (resp.status_code,json.loads(resp.text))

    def run(self):
        """
        任务主函数,获取期货业协会网站中期货公司信息,批量插入数据库
        """
        page_no = 1
        const_pagesize = 20
        arry_comp_link = []
        req_list = "http://www.cfachina.org/orc-report/api/basic/companyBaseInfo"
        self.update_step_struct(['获取期货公司列表','获取具体公司信息'],'running')
        self.update_current_step('获取期货公司列表','running')
        while True:
            pay_load = {
                'pageNo':page_no,
                'pageSize':const_pagesize,
                'agencyName':'',
                'keyword':'',
                'flag':1
            }
            (ret_code,dic_ret) = self.req_json(req_list,pay_load)
            if ret_code == 200:
                if 'data' in dic_ret:
                    arry_comp_link.extend(dic_ret['data']['dataList'])
                    if page_no == 1:
                        page_step = []
                        for page in range(1,int(dic_ret['data']['total']/const_pagesize)+2):
                            page_step.append(f'获取第{page}页')
                        self.update_step_struct(page_step,'running;获取期货公司列表')
                        self.update_current_step(f'获取第{page_no}页','running;获取期货公司列表')
                    else:
                        self.update_current_step(f'获取第{page_no}页','running;获取期货公司列表')
                    if len(arry_comp_link) < dic_ret['data']['total']:
                        page_no +=1
                    else:
                        break
        insert_stmt = insert(FutuID).values(arry_comp_link)
        on_dupkey_stmt = insert_stmt.on_duplicate_key_update(
            organId=insert_stmt.inserted.organId,
            name=insert_stmt.inserted.name
        )
        self.sql_connection.execute(on_dupkey_stmt)
        self.update_current_step('done','')
        return (0,'success',{})

    @classmethod
    def test(cls):
        """
        单元测试
        """
        # self.run()
        return True
