"""
各类工具函数
"""
import os
import random
import tempfile
import uuid
import requests
import time
import json
import datetime
from dateutil.parser import parse


def run_sys_command(cmd) -> tuple:
    """
    执行系统命令,并获取返回值和输出文本
    """
    cmd_code = 0
    cmd_content = ''
    tmp_dir = tempfile.gettempdir()+'/daline/'
    if not os.path.exists(tmp_dir):
        os.mkdir(tmp_dir)
    temp_out = tmp_dir+str(uuid.uuid4())
    run_cmd = cmd+' > '+temp_out
    cmd_code = os.system(run_cmd)
    if not os.path.exists(temp_out):
        return (cmd_code, 'run command failed')
    with open(temp_out, 'r', encoding='utf8') as r_file:  # 打开文件
        cmd_content = r_file.read()  # 读取文件
        r_file.close()
        os.remove(temp_out)
    return(cmd_code, cmd_content)


def req_url(url, header=None) -> tuple:
    """
    请求url,为防止流量限制,随机暂停1-3s

    Parameters
        url: 目标url
        header: 请求的header
    Return
        status_code: 状态码
        text: 内容
    """
    if not header:
        header = {"User-Agent":
                  "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) \
Chrome/72.0.3626.81 Safari/537.36"}
    resp_json = requests.post(url, headers=header, timeout=30)
    time.sleep(random.uniform(1, 3))
    return (resp_json.status_code, resp_json.text)


def req_json(url, header=None) -> tuple:
    """
    请求restful类接口数据

    Parameters
        url: 目标url
        header: 请求的header
    Return
        status_code: 状态码
        text: 内容
    """
    dic_ret = {}
    (status_code, text) = req_url(url, header)
    if status_code == 200:
        dic_ret = json.loads(text)
    return (status_code, dic_ret)

# pylint:disable=broad-except


def check_int(key_item, arry_dic) -> bool:
    """
    检测是类型
    """
    try:
        for item in arry_dic:
            if key_item in item:
                value = int(item[key_item])
                if value >= 2147483647:
                    return False
        return True
    except Exception as _:
        return False


def check_bigint(key_item, arry_dic) -> bool:
    """
    检测是类型
    """
    try:
        for item in arry_dic:
            if key_item in item:
                int(item[key_item])
        return True
    except Exception as _:
        return False


def check_double(key_item, arry_dic) -> bool:
    """
    检测是类型
    """
    try:
        for item in arry_dic:
            if key_item in item:
                float(item[key_item])
        return True
    except Exception as _:
        return False


def check_date(key_item, arry_dic) -> bool:
    """
    检测是类型
    """
    try:
        for item in arry_dic:
            if key_item in item:
                parse(item[key_item])
            if item[key_item].find(':') >= 0 or \
                    item[key_item].find('T') >= 0 or \
                    item[key_item].find(' ') >= 0:
                return False
        return True
    except Exception as _:
        return False


def check_time(key_item, arry_dic) -> bool:
    """
    检测是类型
    """
    try:
        for item in arry_dic:
            if key_item in item:
                parse(item[key_item])
            if item[key_item].find('-') >= 0 or item[key_item].find('/') >= 0 or \
                    item[key_item].find('T') >= 0 or item[key_item].find(' ') >= 0:
                return False
        return True
    except Exception as _:
        return False


def check_datetime(key_item, arry_dic) -> bool:
    """
    检测是类型
    """
    try:
        for item in arry_dic:
            if key_item in item:
                parse(item[key_item])
        return True
    except Exception as _:
        return False


def check_varchar(key_item, arry_dic) -> bool:
    """
    检测是类型
    """
    max_length = 0
    for item in arry_dic:
        if key_item in item and item[key_item]:
            max_length = max(max_length, len(item[key_item]))
            if len(item[key_item]) > 4096:
                return (False, max_length)
    return (True, max_length)


def check_text(key_item, arry_dic) -> bool:
    """
    检测是类型
    """
    max_length = 0
    for item in arry_dic:
        if key_item in item and item[key_item]:
            max_length = max(max_length, len(item[key_item]))
            if len(item[key_item]) > 65535:
                return False
    return True


def check_mediumtext(key_item, arry_dic) -> bool:
    """
    检测是类型
    """
    max_length = 0
    for item in arry_dic:
        if key_item in item and item[key_item]:
            max_length = max(max_length, len(item[key_item]))
            if len(item[key_item]) > 16777215:
                return False
    return True


def utc_shanghai(tz_time: datetime):
    """
    将datetime时间类型转换到UTC+8
    """
    shanghai_tz = datetime.timezone(datetime.timedelta(hours=8))
    return tz_time.astimezone(shanghai_tz)
